/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.part.ViewPart;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.ui.editpart.DiagramEditPartBase;
import org.fortiss.tooling.base.ui.editpart.ElementEditPartBase;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.util.SelectionUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * This class provides a limited set of variables and methods to annotation views. Its main purposes
 * are the management of selections in the UI and their mapping to the corresponding model elements,
 * and maintaining knowledge about the available annotation entries (see {@link AnnotationEntry}).
 * 
 * @author eder
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 40FF672DC52B17062D1716341177D76A
 */
public abstract class AnnotationViewPartBase extends ViewPart implements ISelectionListener,
		IAnnotationViewPart {

	/** the currently selected object */
	private IModelElement currentlySelectedObject;

	/**
	 * Set of {@link IModelElement}s for which the {@link #changeListener} has been installed to
	 * watch for model changes (to trigger an update of this {@link AnnotationViewPartBase}).
	 */
	Set<EObject> watchedModelElements = new HashSet<EObject>();

	/**
	 * {@link Adapter} to watch for the addition, removal or value change of model elements
	 * to/from/in elements visible in this {@link AnnotationViewPartBase} (to trigger according
	 * updates).
	 */
	private final Adapter changeListener = new AdapterImpl() {
		@Override
		public void notifyChanged(Notification notification) {
			if(notification.getEventType() == Notification.ADD ||
					notification.getEventType() == Notification.REMOVE ||
					notification.getEventType() == Notification.SET) {

				Object notifier = notification.getNotifier();
				if((notifier instanceof EObject) &&
						(IPersistencyService.INSTANCE.getTopLevelElementFor((EObject)notifier) != null)) {
					// Ignore update requests for elements that are not hook to a top-level element,
					// e.g. SET events that are fired when a component is removed (before the actual
					// REMOVE event). Otherwise, the AnnotationValueService would create a NPE
					// during the preparation of the context via
					// KernelModelElementUtils.runAsCommand().

					if(notifier instanceof IModelElement) {
						update((IModelElement)notifier);
					} else if(notifier instanceof IAnnotatedSpecification) {
						update(((IAnnotatedSpecification)notifier).getSpecificationOf());
					}
				}
			}
		}
	};

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		currentlySelectedObject = null;
		currentlySelectedObject = SelectionUtils.checkAndPickFirst(selection, IModelElement.class);

		if(currentlySelectedObject == null) {
			ElementEditPartBase<?> editPart =
					SelectionUtils.checkAndPickFirst(selection, ElementEditPartBase.class);
			if(editPart != null && editPart.getModel() instanceof IModelElement) {
				currentlySelectedObject = (IModelElement)editPart.getModel();
			} else {
				// Not all editors are derived from ElementEditPartBase, most notably
				// PlatformArchitectureDiagramEditPart
				DiagramEditPartBase<?> diagramPart =
						SelectionUtils.checkAndPickFirst(selection, DiagramEditPartBase.class);
				if(diagramPart != null && diagramPart.getModel() instanceof IModelElement) {
					currentlySelectedObject = (IModelElement)diagramPart.getModel();
				}
			}
		}
		if(currentlySelectedObject != null) {
			update(currentlySelectedObject);
		}
	}

	/** Update registration of model listeners. */
	private void updateChangeListener(Collection<AnnotationEntry> annotationEntryList) {
		// Determine all root elements
		Set<EObject> rootElements = new HashSet<EObject>();
		for(AnnotationEntry entry : annotationEntryList) {
			EObject rootElement = entry.getModelElement();
			while(!(rootElement instanceof IProjectRootElement)) {
				rootElement = rootElement.eContainer();
			}
			rootElements.add(rootElement);
		}

		// Determine all IModelElements, and their specifications below the current root elements.
		// Specifications must be watched, too, since the state of an editing support for a given
		// specification can depend on the state of one or more other specifications.
		Set<EObject> currentWatchedModelElements = new HashSet<EObject>();
		for(EObject rootElement : rootElements) {
			for(EObject modelElement : EcoreUtils.getChildrenWithType(rootElement,
					IModelElement.class)) {
				currentWatchedModelElements.add(modelElement);
				currentWatchedModelElements.addAll(((IModelElement)modelElement)
						.getSpecifications());
			}
			if(rootElement instanceof IModelElement) {
				currentWatchedModelElements.add(rootElement);
				currentWatchedModelElements
						.addAll(((IModelElement)rootElement).getSpecifications());
			}
		}

		// Add change listeners to IModelElements that are about to appear in the view
		Set<EObject> tmpWatchedModelElements = new HashSet<EObject>(currentWatchedModelElements);
		tmpWatchedModelElements.removeAll(watchedModelElements);
		for(EObject modelElement : tmpWatchedModelElements) {
			modelElement.eAdapters().add(changeListener);
		}

		// Remove change listeners IModelElements that are about to disappear from the view
		tmpWatchedModelElements = new HashSet<EObject>(watchedModelElements);
		tmpWatchedModelElements.removeAll(currentWatchedModelElements);
		for(EObject modelElement : tmpWatchedModelElements) {
			modelElement.eAdapters().remove(changeListener);
		}

		watchedModelElements = currentWatchedModelElements;
	}

	/**
	 * Update the {@link AnnotationViewPartBase}: delegate update of concrete view to
	 * {@link #update(Collection)}, and update model change listeners.
	 */
	private void update(IModelElement modelElement) {
		Set<AnnotationEntry> annotationEntries =
				new TreeSet<AnnotationEntry>(new Comparator<AnnotationEntry>() {
					/** Compare AnnotationEntries by the model element they refer to. */
					@Override
					public int compare(AnnotationEntry a1, AnnotationEntry a2) {
						// Equality of model element implies equal hashCode(), order is arbitrary
						// but not relevant for ensuring uniqueness of annotation entries.
						//
						// int Integer.compare(int x, int y) requires JDK 1.7 which is not available
						// on the build server
						return new Integer(a1.getModelElement().hashCode()).compareTo(a2
								.getModelElement().hashCode());
					}
				});

		// Collect all AnnotationEntries for the current model element and its offspring
		annotationEntries.addAll(IAnnotationValueService.INSTANCE.getValues(modelElement));
		for(IModelElement childModelElement : EcoreUtils.getChildrenWithType(modelElement,
				IModelElement.class)) {
			annotationEntries.addAll(IAnnotationValueService.INSTANCE.getValues(childModelElement));
		}

		// Update the view
		update(annotationEntries);

		updateChangeListener(annotationEntries);
	}

	/** Update concrete view */
	protected abstract void update(Collection<AnnotationEntry> annotationEntries);

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		getSite().getWorkbenchWindow().getSelectionService().addSelectionListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public IModelElement getCurrentlySelectedObject() {
		return currentlySelectedObject;
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		getSite().getWorkbenchWindow().getSelectionService().removeSelectionListener(this);

		for(EObject modelElement : watchedModelElements) {
			modelElement.eAdapters().remove(changeListener);
		}
		watchedModelElements.clear();

		super.dispose();
	}
}
